/* NEC remote infrared RMT example

 This example code is in the Public Domain (or CC0 licensed, at your option.)

 Unless required by applicable law or agreed to in writing, this
 software is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
 CONDITIONS OF ANY KIND, either express or implied.
 */
#include <stdio.h>
#include <string.h>
#include <sys/unistd.h>
#include "freertos/FreeRTOS.h"
#include "freertos/task.h"
#include "freertos/queue.h"
#include "freertos/semphr.h"
#include "esp_err.h"
#include "esp_log.h"
#include "driver/rmt.h"
#include "driver/periph_ctrl.h"
#include "soc/rmt_reg.h"
#include "k2audio.h"
#include "esp_task_wdt.h"



typedef struct {
	uint16_t spaceLong;
	uint16_t spaceShort;
	uint16_t burst;
	uint16_t leadBurst;
	uint16_t leadSpace;
	uint8_t noOfBits;
	char name[8];
	bool verifyCommand;
}NEC_IR_DATA;

NEC_IR_DATA nec_ir;

static const char *TAG = "IR";

bool decodeSonyIR(rmt_item32_t *itemPtr);
bool decodePR_IR(rmt_item32_t *itemPtr);
void decodeModeNEC();
void decodeModeJVC();
void decodeModePANA48();
void decodeModeSSUNG();
uint8_t reverseBits(uint8_t in);

void blastNEC(uint16_t addr, uint8_t cmd);
void print_bin(unsigned int integer);

uint16_t spaceLong, spaceShort, burst, leadBurst, leadSpace, noOfBits;

rmt_item32_t* getDurations(uint32_t* d1, uint32_t* d0, rmt_item32_t* dataPtr);


static uint16_t global_dev_code;
static uint8_t global_cmd_code;

//CHOOSE SELF TEST OR NORMAL TEST
//#define RMT_RX_SELF_TEST   1

//Test with infrared LED, we have to enable carrier for transmitter
//When testing via IR led, the receiver waveform is usually active-low.
#define RMT_RX_ACTIVE_LEVEL  0   /*!< Active low for on board receivers */
#define RMT_TX_CARRIER_EN    1   /*!< Enable carrier for IR transmitter test with IR led */
#define RMT_RX_GPIO_NUM  34     /*!< GPIO number for receiver */
#define RMT_TX_GPIO_NUM  23     /*!< GPIO number for transmitter signal */
#define RMT_TX_CHANNEL    1     /*!< RMT channel for transmitter */
#define RMT_RX_CHANNEL    0     /*!< RMT channel for receiver */
#define RMT_CLK_DIV      100    /*!< RMT counter clock divider */
#define RMT_TICK_10_US    (80000000/RMT_CLK_DIV/100000)   /*!< RMT counter value for 10 us.(Source clock is APB clock) */
#define RMT_TX_DATA_NUM  10    /*!< NEC tx test data number was 100 */
#define rmt_item32_tIMEOUT_US  10000   /*!< RMT receiver timeout value(us) NOTE some NEC based codes have the lead in burst of 9mS */

bool matchIR(BYTE num) {
	WORD error;
	WORD_VAL received_data;
	WORD_VAL stored_data;
	BYTE count;
	rmt_item32_t *rec_ir_ptr;
	rmt_item32_t *ir_data_ptr;

	RunFlags.matchIRinUse = true;

	// rmt items are 2 16 bit word with the msb setting if it is burst or space.
	// so we need to get the values with the masb to compare.

	if (storedPairCount != pairCount) {
		return false;
	}

	rec_ir_ptr = (rmt_item32_t*) rec_IR_data;
	ir_data_ptr = (rmt_item32_t*) rmtItem;
	for (count = 0; count < (pairCount * 2); count++) {

		received_data.Val = rec_ir_ptr[count].duration1;
		received_data.Val = received_data.Val * 10;  // convert to uS from RMT ticks
		received_data.Val = received_data.Val / RMT_TICK_10_US;

		stored_data.Val = ir_data_ptr[count].duration1;
		error = received_data.Val >> 2; // /4 25% window
		if ((stored_data.Val > (received_data.Val + error)) || (stored_data.Val < (received_data.Val - error))) {
			RunFlags.matchIRinUse = false;
			return false;
		}

		received_data.Val = rec_ir_ptr[count].duration0;
		received_data.Val = received_data.Val * 10; // convert to uS from RMT ticks
		received_data.Val = received_data.Val / RMT_TICK_10_US;
		stored_data.Val = ir_data_ptr[count].duration0;
		error = received_data.Val >> 2; // /4
		if ((stored_data.Val > (received_data.Val + error)) || (stored_data.Val < (received_data.Val - error))) {
			RunFlags.matchIRinUse = false;
			return false;
		}
	}

	RunFlags.matchIRinUse = false;
	return true;
}

/*
 * @brief RMT transmitter initialization
 */
static void IR_tx_init() {
	rmt_config_t rmt_tx;
	rmt_tx.channel = RMT_TX_CHANNEL;
	rmt_tx.gpio_num = RMT_TX_GPIO_NUM;
	rmt_tx.mem_block_num = 1;
	rmt_tx.clk_div = RMT_CLK_DIV;
	rmt_tx.tx_config.loop_en = false;
	rmt_tx.tx_config.carrier_duty_percent = 50;
	rmt_tx.tx_config.carrier_freq_hz = 38000;
	rmt_tx.tx_config.carrier_level = 1;
	rmt_tx.tx_config.carrier_en = RMT_TX_CARRIER_EN;
	rmt_tx.tx_config.idle_level = 0;
	rmt_tx.tx_config.idle_output_en = true;
	rmt_tx.rmt_mode = 0;
	rmt_config(&rmt_tx);
	rmt_driver_install(rmt_tx.channel, 0, 0);
}

/*
 * @brief RMT receiver initialization
 */
static void IR_rx_init() {
	rmt_config_t rmt_rx;
	rmt_rx.channel = RMT_RX_CHANNEL;
	rmt_rx.gpio_num = RMT_RX_GPIO_NUM;
	rmt_rx.clk_div = RMT_CLK_DIV;
	rmt_rx.mem_block_num = 1;
	rmt_rx.rmt_mode = RMT_MODE_RX;
	rmt_rx.rx_config.filter_en = true;
	rmt_rx.rx_config.filter_ticks_thresh = 100;
	rmt_rx.rx_config.idle_threshold = rmt_item32_tIMEOUT_US / 10 * (RMT_TICK_10_US);
	rmt_config(&rmt_rx);
	rmt_driver_install(rmt_rx.channel, 1000, 0);

	RunFlags.debug = true;

}

/**
 * @brief RMT receiver demo, this task will print each received NEC data.
 *
 */
void IR_rx_task() {

	int channel = RMT_RX_CHANNEL;
	IR_rx_init();
	RingbufHandle_t rb = NULL;
	//get RMT RX ringbuffer
	rmt_get_ringbuf_handle(channel, &rb);
	rmt_rx_start(channel, 1);
	while (rb) {
		size_t rx_size = 0;
		//try to receive data from ringbuffer.
		//RMT driver will push all the data it receives to its ringbuffer.
		//We just need to parse the value and return the spaces of ringbuffer.
		rmt_item32_t *itemPtr = (rmt_item32_t*) xRingbufferReceive(rb, &rx_size, 1000);
		rmt_item32_t *dataPtr = itemPtr;

		if (itemPtr) {

			ESP_LOGI(TAG, "IR data size = %d received", rx_size);
			if (rx_size > 0x04) {
				gpio_set_level(STATUS_GPIO, 1);
				pairCount = rx_size / 4;
				frequency = 36;
				esp_task_wdt_reset();

				//decodeSonyIR(dataPtr);
				decodeModeNEC();
				//decodeModeJVC();
				//decodeModePANA48();
				//decodeModeSSUNG();
				decodePR_IR(dataPtr);
				rmtToAscii(dataPtr, 0);
				//setData2Ready();
			}
			//after parsing the data, return spaces to ringbuffer.
			vRingbufferReturnItem(rb, (void*) itemPtr);

		}

	}
	vTaskDelete(NULL);
}


rmt_item32_t* getDurations(uint32_t* d1, uint32_t* d0, rmt_item32_t* dataPtr){
// get burst and space times in uS

	*d0 = (int) dataPtr->duration0;
	*d1 = (int) dataPtr->duration1;
	*d0 = *d0 * 10;
	*d1 = *d1 * 10;

	*d0 = *d0 / RMT_TICK_10_US;
	*d1 = *d1 / RMT_TICK_10_US;
	dataPtr++;

	return dataPtr;

}

bool decodeSonyIR(rmt_item32_t *itemPtr) {
	uint16_t T1L, T1S, T2L, T2S;
	uint16_t errOne, errTwo;
	uint32_t d0 = 0, d1 = 0;

	bool rtn = true;
	uint8_t bitCount = 0;
	uint32_t data = 1;
	uint32_t result = 0;
	uint16_t command = 0, device = 0, xtra = 0;

	if (pairCount < 12) {
		return false;
	}

	errOne = 1200 >> 2; //  div by 4
	errTwo = 600 >> 2;

	// window for dataOne
	T1L = 1200 + errOne;
	T1S = 1200 - errOne;

	// window for dataTwo
	T2L = 600 + errTwo;
	T2S = 600 - errTwo;
	itemPtr = getDurations(&d1, &d0, itemPtr);


	if (d0 < 2000 || d0 > 2800) {
		return false;
	}

	if (d1 < 400 || d1 > 900) {
		return false;
	}

	for (bitCount = 0; bitCount <pairCount - 2; bitCount++) {

		itemPtr = getDurations(&d1, &d0, itemPtr);

		if (d0 == 0x00) {
			bitCount = pairCount;
		}
		if (d1 == 0x00) {
			d1 = 0x2000;

		}

		if ((T1S < d0) && (T1L > d0)) {
			result = result | data;
		} else if ((T2S < d0) && (T2L > d0)) {
			// valid 0
		} else {
			// not valid code
			rtn = false;
			break;
		}
		data = data << 1;

		if ((T2S < d1) && (T2L > d1)) {
			// OK
		} else {
			if (d1 != 0x2000) {
				rtn = false;
			}
			break;
		}

	}


	if (rtn == false) {
		ESP_LOGI(TAG, "Return is false");
		//	sprintf(debugInfo,"bitCount = %d pairCount = %d and data = %s", bitCount, pairCount, out);
		return rtn;
	}
	bitCount++;


	// bitcount should be 12, 15 or 20
	// the LSB of command will now be in the MSB of out

	if (bitCount == 12) {
		command = result & 0x7F;
		result = result>> 7;
		device = result & 0x1F;
	} else if (bitCount == 15) {
		command = result & 0x7F;
		result = result>> 7;
		device = result &0xFF;
	} else if (bitCount == 20) {
		command = result & 0x7F;
		result = result>> 7;
		device = result &0xFF;
		result = result >> 8;
		xtra = result;
	} else {
		return false;

	}

	sprintf(debugInfo, "Sony %d bit \n Cmd = %d\n Dev = %d + xtra %d", bitCount, command, device, xtra);
	RunFlags.displayUpdated = true;

	return rtn;
}



/*****Pacific rim IR****from various Internet sources NOT verified *********************************
 *  name				Lead burst	lead space		burst time 	space 1		space 0	     no bits
 *
 * *************************************************************************************************
 *
 *
 * for those using 38Khz IR the 526 1540 timings look good at exactly 20 cycles for the burst
 * and 60 for a long space
 *
 * Panasonic old		3600		1700(3600) 		900 		2700		900				22
 *
 * Panasonic new		3600		1700 			440			1250		440				48
 *
 * Sanyo				3380		1690			420			1260		420				48 same as New pana  ??
 *
 * Sharp 				----		----			320			1680		680				15
 *
 * Toshiba				9000	 	4500			552			1656		552				32
 *
 * Denon				----		----			275			1900		775				15
 *
 * RCA					4000		4000			500			2000		1000			24
 *
 * NEC					9000		4500			560			1680		560				32
 *
 * Samsung				4500		""""			""""		""""		""""			32
 *
 * JVC					8400		4200			526			1540		526				16 - have seen some JVC with  lead in burst / space only on the first code sent
 *
 * Daewoo				8000		4000			550			1450		550				14
 *
 * RECS80				158			5060			158			7590		5060			11
 *
 ***********************************************************************************************/
void decodeModeNEC(){

	nec_ir.leadBurst = 9000;
	nec_ir.leadSpace = 4500;
	nec_ir.spaceLong = 1680;
	nec_ir.spaceShort = 560;
	nec_ir.burst = 560;
	nec_ir.noOfBits = 32;
	strcpy(nec_ir.name,"NEC");
	nec_ir.verifyCommand = true;


}

void decodeModeSSUNG(){

	nec_ir.leadBurst = 4500;
	nec_ir.leadSpace = 4500;
	nec_ir.spaceLong = 1680;
	nec_ir.spaceShort = 560;
	nec_ir.burst = 560;
	nec_ir.noOfBits = 32;
	strcpy(nec_ir.name,"SSUNG");
	nec_ir.verifyCommand = true;


}



void decodeModeJVC(){

	nec_ir.leadBurst = 8400;
	nec_ir.leadSpace = 4200;
	nec_ir.spaceLong = 1540;
	nec_ir.spaceShort = 526;
	nec_ir.burst = 526;
	nec_ir.noOfBits = 16;
	strcpy(nec_ir.name,"PANA48");
	nec_ir.verifyCommand = false;


}

void decodeModePANA48(){

	nec_ir.leadBurst = 3600;
	nec_ir.leadSpace = 1700;
	nec_ir.spaceLong = 1250;
	nec_ir.spaceShort = 440;
	nec_ir.burst = 440;
	nec_ir.noOfBits = 48;
	strcpy(nec_ir.name,"JVC");
	nec_ir.verifyCommand = false;

}

uint8_t reverseBits(uint8_t in){
	uint8_t out = 0, x, mask;
	mask = 0x80;
	for (x = 0;x<8; x++){
		out = out  >> 1;
		if (in & mask){
			out = out | 0x80;
		}
		mask = mask >>1;

	}

return out;
}



void print_bin(unsigned int integer)
{
    int i = CHAR_BIT * sizeof integer; /* however many bits are in an integer */
    while(i--) {
        putchar('0' + ((integer >> i) & 1));
    }
    printf("\r\n");
}



bool decodePR_IR(rmt_item32_t *itemPtr) {

	uint16_t T1L, T1S, T2L, T2S;
	uint16_t errOne, errTwo, errLB, errLS;
	uint32_t d0 = 0, d1 = 0;

	bool rtn = true;
	uint8_t bitCount = 0, tmp = 0, tmp2 = 0;
	uint32_t result = 0;
	uint16_t command = 0, commandI = 0, device = 0, xtnded =0, commandR = 0;
	uint16_t manuCode =0;
	ESP_LOGI(TAG,"NEC decode started");


	if (pairCount < nec_ir.noOfBits) {
		ESP_LOGI(TAG," error pairCount = %d noOfBits = %d", pairCount, nec_ir.noOfBits);
		return false;
	}

	errOne = nec_ir.spaceLong >> 2; //  div by 4
	errTwo = nec_ir.spaceShort >> 2;

	errLB = nec_ir.leadBurst >> 2;
	errLS = nec_ir.leadSpace >> 2;

	// window for dataOne
	T2L =  nec_ir.spaceLong + errOne;
	T2S = nec_ir.spaceLong- errOne;

	// window for dataTwo
	T1L = nec_ir.spaceShort + errTwo;
	T1S = nec_ir.spaceShort- errTwo;

	itemPtr = getDurations(&d1, &d0, itemPtr);

	if (d0 < (nec_ir.leadBurst - errLB) || d0 > (nec_ir.leadBurst + errLB)) {
		ESP_LOGI(TAG,"Lead burst error");
		return false;
	}


	if (d1 < (nec_ir.leadSpace - errLS) || d1 > (nec_ir.leadSpace + errLS)) {
		ESP_LOGI(TAG,"Lead space error");
		return false;
	}

	for (bitCount = 0; bitCount <pairCount - 2; bitCount++) {

		itemPtr = getDurations(&d1, &d0, itemPtr);

		if (d0 == 0x00) {
			bitCount = pairCount;
		}
		if (d1 == 0x00) {
			d1 = 0x2000;

		}

		if ((d0 <T1S) && (d0 > T1L)) {
			// burst always 560 or ??
			ESP_LOGI(TAG,"burst error %d", d0);
			rtn = false;
			break;
		}

		result = result << 1;
		if(bitCount < 16){
			// first 16 bits for 48 bit codes
			manuCode = manuCode << 1;
		}
		if ((T2S < d1) && (T2L > d1)) {
			// OK '1'
			result = result + 1;
			if(bitCount < 16){
				// first 16 bits for 48 bit codes
				manuCode = manuCode +1;
			}
		} else if ((T1S < d1) && (T1L > d1)) {
			// 560 us for a 0
		} else {
			if (d1 != 0x2000) {
				rtn = false;
			}
			ESP_LOGI(TAG,"Space error %d", d1);
			break;
		}


	}


	if (rtn == false) {
		ESP_LOGI(TAG, "Return is false");
		//	sprintf(debugInfo,"bitCount = %d pairCount = %d and data = %s", bitCount, pairCount, out);
		return rtn;
	}

	if ( bitCount != nec_ir.noOfBits){
		return false;
	}


//	ESP_LOGI(TAG,"Bitcount = %d result %d", bitCount, result);
//	print_bin(result);

	if (nec_ir.noOfBits == 16){
		command = result & 0xFF;
		commandR = reverseBits(command);

		device = result >> 8;
		device = device &0xFF;

		device = reverseBits(device);
		ESP_LOGI(TAG,"Command =  0x%x  Device = 0x%x", commandR, device);
	} else if (nec_ir.noOfBits == 32) {
		command = result & 0xFF;
		commandI = result >> 8;
		commandI = commandI & 0xFF;
		commandR = reverseBits(commandI);

		device = result >> 16;
		xtnded = device >> 8;
		device = device & 0xFF;
		device = reverseBits(device);
		xtnded = reverseBits(xtnded);
		xtnded = xtnded << 8;
		xtnded = xtnded  + device;
		ESP_LOGI(TAG,"Command =  0x%x  extended = 0x%x", commandR, xtnded);
	} else if (nec_ir.noOfBits == 48){
		// this handles just the lower 32 bits of the 48
		command = result & 0xFF;
		commandI = result >> 8;
		commandI = commandI & 0xFF;
		commandR = reverseBits(commandI);

		device = result >> 16;
		xtnded = device >> 8;
		device = device & 0xFF;
		device = reverseBits(device);
		xtnded = reverseBits(xtnded);
		xtnded = xtnded << 8;
		xtnded = xtnded  + device;

		// sort the 16 bit manufacturer code
		tmp = manuCode & 0xFF;
		tmp2 = manuCode >> 8;
		tmp2 = tmp2 & 0xFF;
		tmp = reverseBits(tmp);
		tmp2 = reverseBits(tmp2);
		manuCode = tmp2 << 8 ;
		manuCode = manuCode + tmp;

		ESP_LOGI(TAG,"Command =  0x%x  extended = 0x%x", commandR, xtnded);
		ESP_LOGI(TAG,"Manufacturer's code = 0x%x", manuCode);

	}else {
		ESP_LOGI(TAG,"bitCount error bitCount = %d", bitCount);
		return false;

	}
	if (nec_ir.verifyCommand){
		if ((command  + commandI) != 0xFF){
			return false;
		}
	}
	sprintf(debugInfo, "Type %s %d bit \n Cmd = 0x%x\n Dev = 0x%x", nec_ir.name,bitCount, commandR, xtnded);
	RunFlags.displayUpdated = true;

	global_dev_code = xtnded;
	global_cmd_code = commandR;

	return rtn;
}




void rmtToAscii(rmt_item32_t *itemPtr, int pos) {
	bool code_acquired = false;
	uint32_t d0, d1;
	char num[5];
	char tmp[2];

	udp_send_data[pos] = 0x4B;
	udp_send_data[pos + 1] = ' ';

	if (frequency == 0x00) {
		frequency = 0x24; // default
	}

	if (pairCount == 0x00) {
		pairCount = 0x30; // pair count 0x30 default value
	}

	byteToAsciiHex(frequency, tmp);
	udp_send_data[pos + 2] = tmp[0];
	udp_send_data[pos + 3] = tmp[1];
	byteToAsciiHex(pairCount, tmp);
	udp_send_data[pos + 4] = tmp[0];
	udp_send_data[pos + 5] = tmp[1];
	udp_send_data[pos + 6] = 0x20; // space;

	int x = 7;
	if (RunFlags.debug) {
		printf("K ");
		printf("%02X%02X", frequency, pairCount);
	}

	for (int z = 0; z < pairCount; z++) {
		if (!code_acquired) {

			itemPtr = getDurations(&d1, &d0, itemPtr);

			//	if ((d1 == 0x00) || ((d1 >= 0x1F00) || (d0 >= 0x1F00))) { // need to ad z = pairCount-1;
			if ((d1 == 0x00) || (z == (pairCount - 1))) {
				d1 = 0x2000;
			}

			if (RunFlags.debug) {

				printf(" %04X", d0);
				printf(" %04X", d1);
			}

			sprintf(num, "%04X", d0);
			udp_send_data[pos + x] = num[0];
			udp_send_data[pos + x + 1] = num[1];
			udp_send_data[pos + x + 2] = num[2];
			udp_send_data[pos + x + 3] = num[3];
			udp_send_data[pos + x + 4] = 0x20;

			sprintf(num, "%04X", d1);
			udp_send_data[pos + x + 5] = num[0];
			udp_send_data[pos + x + 6] = num[1];
			udp_send_data[pos + x + 7] = num[2];
			udp_send_data[pos + x + 8] = num[3];
			udp_send_data[pos + x + 9] = 0x20;
			x += 10;

			if (z == pairCount - 1) {
				code_acquired = true;
				udp_data_length = pos + x + 7;
				if (RunFlags.debug) {
					printf(" \r\n");
				}
				z = pairCount;

			}
		}

	}

}

/**
 * @brief RMT transmitter demo, )
 *
 */
void IR_tx_task() {
	bool localDataReady = false;
	vTaskDelay(10);
	IR_tx_init();
	int channel = RMT_TX_CHANNEL;
	RingbufHandle_t rb = NULL;
//get RMT RX ringbuffer
	rmt_get_ringbuf_handle(channel, &rb);
	pairCount = 16; // set to max
	RunFlags.IR_TX_dataReady = false;

	rmt_item32_t *itemPtr;
	for (;;) {

		if (RunFlags.IR_TX_dataReady) {
			// ir was received over udp
			RunFlags.IR_TX_dataReady = false;

			asciitoRmt(itemPtr);
			if (pairCount == 0x55){
				uint8_t cmd;
				uint16_t dev;
				WORD_VAL MyWord;
				MyWord.v[1] = ascii2hex(AsciiData[7], AsciiData[6]);
				MyWord.v[0] = ascii2hex(AsciiData[9], AsciiData[8]);
				cmd = ascii2hex(AsciiData[11], AsciiData[10]);
				dev = MyWord.Val;
				decodeModeNEC();
				blastNEC(dev, cmd);
			} else {
				itemPtr = rmtItem;
				global_dev_code = 0x00;
				//decodeSonyIR(itemPtr);
				decodeModeNEC();
				//decodeModeJVC();
				//decodeModePANA48();
				decodePR_IR(itemPtr);

			}

			//localDataReady = true;
		}

		if (localDataReady) {
			itemPtr = rmtItem;
			localDataReady = false;
			//To send data according to the waveform items.
			rmt_write_items(channel, itemPtr, pairCount, true);
			//Wait until sending is done.
			rmt_wait_tx_done(channel, portMAX_DELAY);
			//before we free the data, make sure sending is already done.
			//	itemPtr = startPtr;

			// if done clear the memory
			//memset((void*) itemPtr, 0, size);
		}
		vTaskDelay(100 / portTICK_PERIOD_MS);
	}
//	free(itemPtr);
	vTaskDelete(NULL);

}

void asciitoRmt(rmt_item32_t *itemPtr) {

// SKY BOX UK	POWER:
//	K 2417 0A68 0378 01BC 01BC 01BC 01BC 01BC 0378 01BC 0378 0378 01BC 01BC 01BC 01BC 01BC 01BC 01BC 01BC 01BC 0378 0378 0378 01BC 01BC 01BC 01BC 0378 01BC 01BC 01BC 01BC 01BC 01BC 01BC 01BC 01BC 01BC 0378 01BC 01BC 0378 01BC 01BC 01BC 2000

	WORD_VAL MyWord;

	frequency = ascii2hex(AsciiData[3], AsciiData[2]);
	pairCount = ascii2hex(AsciiData[5], AsciiData[4]);

	ESP_LOGI(TAG, "ascii to rmt %c %d %d ", AsciiData[0], frequency, pairCount);

	int z = 7; // assumes ascii including the header K xxxx
	for (int pair = 1; pair < pairCount + 1; pair++) {

		MyWord.v[1] = ascii2hex(AsciiData[z + 1], AsciiData[z]);
		MyWord.v[0] = ascii2hex(AsciiData[z + 3], AsciiData[z + 2]);

		MyWord.Val = (MyWord.Val / 10) * RMT_TICK_10_US;

		itemPtr->duration0 = MyWord.Val;
		itemPtr->level0 = 1;
		z = z + 5;

		MyWord.v[1] = ascii2hex(AsciiData[z + 1], AsciiData[z]);
		MyWord.v[0] = ascii2hex(AsciiData[z + 3], AsciiData[z + 2]);

		MyWord.Val = (MyWord.Val / 10) * RMT_TICK_10_US;
		itemPtr->duration1 = MyWord.Val;
		itemPtr->level1 = 0;
		itemPtr++;

		z = z + 5;

	}
//ESP_LOGI(TAG, "ascii complete");

}


void blastNEC(uint16_t addr, uint8_t cmd){
	// this is just NEC at the moment
	uint8_t count, bar_cmd;
	uint8_t tmp;
	bar_cmd = cmd;
	WORD_VAL MyWord;
	rmt_item32_t *itemPtr;
	int channel = RMT_TX_CHANNEL;

	ESP_LOGI(TAG,"bLAST ir-> addr = 0x%x cmd = 0x%x",addr, cmd);
	itemPtr = rmtItem;
	MyWord.Val = nec_ir.leadBurst;
	MyWord.Val = (MyWord.Val / 10) * RMT_TICK_10_US;
	itemPtr->duration0 = MyWord.Val;
	itemPtr->level0 = 1;

	MyWord.Val = nec_ir.leadSpace;
	MyWord.Val = (MyWord.Val / 10) * RMT_TICK_10_US;
	itemPtr->duration1 = MyWord.Val;
	itemPtr->level1 = 0;
	itemPtr++;

	tmp = addr >> 8;


	for (count = 0; count < 8; count++){
		MyWord.Val = nec_ir.burst;
		MyWord.Val = (MyWord.Val / 10) * RMT_TICK_10_US;
		itemPtr->duration0 = MyWord.Val;
		itemPtr->level0 = 1;
		if (tmp & 0x01){
			MyWord.Val = nec_ir.spaceLong;
		}else {
			MyWord.Val = nec_ir.spaceShort;
		}
		MyWord.Val = (MyWord.Val / 10) * RMT_TICK_10_US;
		itemPtr->duration1 = MyWord.Val;
		itemPtr->level1 = 0;
		itemPtr++;
		tmp = tmp >> 1;
	}

	tmp = addr & 0xFF;


	for (count = 0; count < 8; count++){
		MyWord.Val = nec_ir.burst;
		MyWord.Val = (MyWord.Val / 10) * RMT_TICK_10_US;
		itemPtr->duration0 = MyWord.Val;
		itemPtr->level0 = 1;
		if (tmp & 0x01){
			MyWord.Val = nec_ir.spaceLong;
		}else {
			MyWord.Val = nec_ir.spaceShort;
		}
		MyWord.Val = (MyWord.Val / 10) * RMT_TICK_10_US;
		itemPtr->duration1 = MyWord.Val;
		itemPtr->level1 = 0;
		itemPtr++;
		tmp = tmp >> 1;
	}



	for (count = 0; count < 8; count++){
		// normal
		MyWord.Val = nec_ir.burst;
		MyWord.Val = (MyWord.Val / 10) * RMT_TICK_10_US;
		itemPtr->duration0 = MyWord.Val;
		itemPtr->level0 = 1;
		if (cmd & 0x01){
			MyWord.Val = nec_ir.spaceLong;
		}else {
			MyWord.Val = nec_ir.spaceShort;
		}
		MyWord.Val = (MyWord.Val / 10) * RMT_TICK_10_US;
		itemPtr->duration1 = MyWord.Val;
		itemPtr->level1 = 0;
		cmd = cmd >> 1;
		itemPtr++;

	}


	for (count = 0; count < 8; count++){
		//inverted
		MyWord.Val = nec_ir.burst;
		MyWord.Val = (MyWord.Val / 10) * RMT_TICK_10_US;
		itemPtr->duration0 = MyWord.Val;
		itemPtr->level0 = 1;
		if (bar_cmd & 0x01){
			MyWord.Val = nec_ir.spaceShort;
		}else {
			MyWord.Val = nec_ir.spaceLong;
		}
		MyWord.Val = (MyWord.Val / 10) * RMT_TICK_10_US;
		itemPtr->duration1 = MyWord.Val;
		itemPtr->level1 = 0;
		bar_cmd = bar_cmd >> 1;
		itemPtr++;

		}
	// final bit
		MyWord.Val = nec_ir.burst;
		MyWord.Val = (MyWord.Val / 10) * RMT_TICK_10_US;
		itemPtr->duration0 = MyWord.Val;
		itemPtr->level0 = 1;

		MyWord.Val = nec_ir.spaceLong;
		MyWord.Val = (MyWord.Val / 10) * RMT_TICK_10_US;
		itemPtr->duration1 = MyWord.Val;
		itemPtr->level1 = 0;
		itemPtr++;


		itemPtr = rmtItem;


		//To send data according to the waveform items.
		rmt_write_items(channel, itemPtr, pairCount, true);
		//Wait until sending is done.
		rmt_wait_tx_done(channel, portMAX_DELAY);

}


