/* SPIFFS filesystem example.
 This example code is in the Public Domain (or CC0 licensed, at your option.)

 Unless required by applicable law or agreed to in writing, this
 software is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
 CONDITIONS OF ANY KIND, either express or implied.
 */

#include <stdio.h>
#include <string.h>
#include <sys/unistd.h>
#include <sys/stat.h>
#include "esp_err.h"
#include "esp_log.h"
#include "esp_spiffs.h"
#include "k2audio.h"

bool writeSpiffsFileData(char* fileName, uint8_t* data, uint16_t length);
bool readSpiffsFileData(char* fileName, uint8_t* data, uint16_t length);

static const char *TAG = "SPIFFS";


void init_spiffs(void) {
	ESP_LOGI(TAG, "Initializing SPIFFS");

	esp_vfs_spiffs_conf_t conf = { .base_path = "/spiffs", .partition_label =
	NULL, .max_files = MAX_FILES, .format_if_mount_failed = true };
	// Use settings defined above to initialize and mount SPIFFS filesystem.

	// Note: esp_vfs_spiffs_register is an all-in-one convenience function.
	esp_err_t ret = esp_vfs_spiffs_register(&conf);

	if (ret != ESP_OK) {
		if (ret == ESP_FAIL) {
			ESP_LOGI(TAG, "Failed to mount or format filesystem");
		} else if (ret == ESP_ERR_NOT_FOUND) {
			ESP_LOGI(TAG, "Failed to find SPIFFS partition");
		} else {
			ESP_LOGI(TAG, "Failed to initialize SPIFFS (%s)", esp_err_to_name(ret));
		}
		return;
	}

	size_t total = 0, used = 0;
	ret = esp_spiffs_info(NULL, &total, &used);
	if (ret != ESP_OK) {
		ESP_LOGI(TAG, "Failed to get SPIFFS partition information (%s)", esp_err_to_name(ret));
	} else {
		ESP_LOGI(TAG, "Partition size: total: %d, used: %d", total, used);
	}

}

bool writeSpiffsFileData(char* fileName, uint8_t* data, uint16_t length) {

	ESP_LOGI(TAG, "Opening file for writing");
	FILE* f = fopen("/spiffs/AppValues.bin", "w");
	if (f == NULL) {
		ESP_LOGI(TAG, "Failed to open file for writing");

		return false;
	}

	int noOfBytes;

	fseek(f, 0x00, SEEK_SET);

	noOfBytes = fwrite(data, 1, length, f);

	fclose(f);

	if (noOfBytes == length) {
		ESP_LOGI(TAG, "Good - bytes written %d ", noOfBytes);
		return true;
	} else {
		ESP_LOGI(TAG, "Fail - bytes written %d required %d", noOfBytes, length);
		return false;
	}

}


bool loadAppValues() {
	bool dataGood = readSpiffsFileData("AppValues.bin", (uint8_t*) &AppValues, sizeof(AppValues));
	xorit(AppValues.wifiPass); // decrypt SSID and PASSWORD
	xorit(AppValues.wifiSSID);
	ESP_LOGI(TAG, "load Decrypted = SSID %s and PASS %s", AppValues.wifiSSID, AppValues.wifiPass);
	return dataGood;
}

bool saveAppValues() {
	uint8_t saveCount;
	bool goodSave;
	APP_VALUES AppCopy;
	memcpy(&AppCopy, &AppValues, sizeof(AppValues));

	xorit(AppCopy.wifiPass); // VERY BASIC encrypt SSID and PASSWORD
	xorit(AppCopy.wifiSSID);
	saveCount =0;

	RunFlags.saveInProgress = true;
	RunFlags.appValuesChanged = false;
	ESP_LOGI(TAG, "save encrypted = SSID %s and PASS %s", AppValues.wifiSSID, AppValues.wifiPass);
	goodSave = writeSpiffsFileData("AppValues.bin", (uint8_t*) &AppCopy, sizeof(AppValues));
	while (!goodSave) {
		if (saveCount > 3) {
			RunFlags.saveInProgress = false;
			return false;
		}
		goodSave = writeSpiffsFileData("AppValues.bin", (uint8_t*) &AppCopy, sizeof(AppValues));
		saveCount++;
	}
	RunFlags.saveInProgress = false;
	return true;
}

bool readSpiffsFileData(char* fileName, uint8_t* data, uint16_t length) {
	FILE* f = fopen("/spiffs/AppValues.bin", "r");
	if (f == NULL) {
		ESP_LOGE(TAG, "Failed to open file for reading");
		return false;
	}

	int noOfBytes = fread(data, 1, length, f);
	ESP_LOGI(TAG, "bytes read %d and bytes expected %d \r\n", noOfBytes, length);

	fclose(f);
	// ESP_LOGI(TAG, "File read");

	if (noOfBytes == length) {
		return true;
	} else {
		return false;
	}

}

void xorit(char* str) {
	uint8_t MAC[6];
	int x = 0;
	memcpy(MAC, AppValues.wifiMAC, 6);
	for (int i = 0; i < strlen(str); i++) {
		str[i] ^= MAC[x];
		x++;
		if (x > 5) {
			x = 0;
		}
	}
}
